; -*-Emacs-Lisp-*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         mh-refile.el
; RCS:          $Header: mh-refile.el,v 1.2 93/01/20 11:13:34 darrylo Exp $
; Description:  Folder refile support routines
; Author:       Darryl Okahata
; Created:      Tue Sep 24 18:33:11 1991
; Modified:     Mon Jan 18 14:09:30 1993 (Darryl Okahata) darrylo@hpsrdmo
; Language:     Emacs-Lisp
; Package:      N/A
; Status:       Experimental
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; auto refile/Fcc folder support.
;; (From: Andy Norman <ange@hplb.hpl.hp.com>, enhanced by Darryl Okahata
;;  <darrylo@sr.hp.com>)
;;
;; If mh-auto-folder-alist is `nil', mh-refile-msg will act as it has
;; always acted.  If mh-auto-folder-alist is set, auto-refile support
;; will be used (see docs for mh-auto-folder-alist for more info).
;;
;; To get quickly started using these routines, add the following lines
;; to your .emacs file:
;;
;;	(setq mh-auto-folder-alist '(
;;				     (t
;;					(mh-get-default-folder . "")
;;				     ))
;;
;; This will cause a "reasonable" default refile/Fcc folder name to be
;; presented to you.  Later, you can change this to something more
;; complex, like:
;;
;;	(setq mh-auto-folder-alist
;;	  '(
;;	     ("From:"
;;	       ("ug8767ea" . "+cretins")
;;	       ("joe blow" . "+joey")
;;	       ("Erik the Viking" . "+cretins")
;;	       ("Erik the Red" . "+cretins")
;;	       ("Erik the Cretin" . "+cretins")
;;	     )
;;	     ("Cc:"
;;	       ("gnu-emacs" . "+gnu")
;;	       ("g\\+\\+" . "+gcc")
;;	       ("gcc" . "+gcc")
;;	       ("gwm" . "+gwm")
;;	     )
;;	     (t
;;	       (mh-get-default-folder . "")
;;	     )
;;	   )
;;	)
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(require 'mh-e)			;; mh-e.el must be loaded before this one.
				;; Note that mh-db.el is also loaded by mh-e.el
(provide 'mh-refile)


(defvar mh-auto-folder-alist nil
  "*Non-nil value should be an alist that will be used to choose a default
folder name when messages are refiled.  The alist should contain elements of
the form:

\(HEADER-NAME
   (REGEXP . FOLDER-NAME) ...
  ...)

or of the form:

\(t
   (FUNCTION . FOLDER-NAME) ...
  ...)

where HEADER-NAME, REGEXP and FOLDER-NAME are strings.

If any part of the contents of the message header named by HEADER-NAME
is matched by the regular expression REGEXP, the corresponding
FOLDER-NAME will be used as the default when prompting for a folder to
refile the message in.  Note that any FOLDER-NAME must have a `+'
prepended to it.


FUNCTION is a name of a function to call (no arguments are passed to
it), with the current buffer being the mh-show-buffer.  If the function
result is not a string and is non-nil (e.g., `t'), the corresponding
FOLDER-NAME will be used.  If the result is a string, the FOLDER-NAME
will be ignored, and the result will instead be used as the folder name.
Note that any string result must have a `+' prepended to it.

Matching is case sensitive.")


(defun mh-get-from-field ()
  "Return the contents of the From: field."
  (let ()
    (save-window-excursion
      (save-excursion
	(mh-show (mh-get-msg-num t))
	(set-buffer mh-show-buffer)
	(mh-get-field "From:")
	))))
  

(defun mh-get-folder-sender-field ()
  "Get the contents of the field to use to determine the default folder.
The following fields are searched, in the order given:

	Reply-To:
	From:
	To:

The first non-empty field found is the one whose contents are returned.
NIL is returned if a non-empty field could not be found."
  (let (field)
    (cond
     ( (not (string= (setq field (mh-get-field "Reply-To:")) ""))
       field)
     ( (not (string= (setq field (mh-get-field "From:")) ""))
       field)
     ( (not (string= (setq field (mh-get-field "To:")) ""))
       field)
     (t nil)
     )))


(defun mh-get-default-folder ()
  "Scan the current MH message and return a default folder name for it.
This function assumes that the current buffer contains an MH message.

It also doesn't conform to RFC822 standards, but it'll do for most
cases.
"
  ;;
  ;; Regarding RFC822 non-comformance:
  ;;
  ;; Properly parsing the Reply-To: header is a real pain and may not be
  ;; possible.  What do you do in the following case?  What should the
  ;; default folder name be?
  ;;
  ;;	   From:     George Jones <Jones@Host.Net>
  ;;	   Sender:   Jones@Host
  ;;	   Reply-To: The Committee: Jones@Host.Net,
  ;;				    Smith@Other.Org,
  ;;				    Doe@Somewhere-Else;
  ;;
  ;; Perhaps, in this case, we should just fall back to the "From:"
  ;; header?
  ;;
  ;; Also, parsing the following address is another pain:
  ;;
  ;;	   Wilt . (the	Stilt) Chamberlain@NBA.US
  ;;
  ;; (The default folder name should be parsed into "Wilt.Chamberlain".)
  ;;
  ;; (All above examples were taken from RFC822.)
  ;;
  (let (from result)
    (if (setq from (mh-get-folder-sender-field))
	(setq result
	      (cond
	       ;; The following is ORDER-DEPENDENT!
	       ( (or (string-match "[ \t]*[^<]+<\\([^/!%@> \t,]+\\)" from)
		     (string-match "[ \t]*\\([^/!%@< \t,]+\\)" from)
		     )
		 (if (not (string= (setq from (substring from
							 (match-beginning 1)
							 (match-end 1)))
				   ""))
		     (concat "+" from)
		   )
		 )
	       ;; We can add more conditionals here to make the scanning more
	       ;; intelligent.
	       ( t
		 ())
	       )))
    (if (not result)
	(if (and mh-last-destination
		 (eq (car mh-last-destination) 'refile))
	    (setq result (cdr mh-last-destination))))
    result
    ))


(defun mh-auto-select-folder (alist operation)
  "Return the name of a folder to refile the current message in.
The current buffer is scanned for this information."
  (let ((mh-current-folder mh-current-folder)
	(mh-showing (and (boundp 'mh-showing) mh-showing))
	case-fold-search
	header
	result)
    (save-window-excursion
      (save-excursion
	(catch 'match
	  (mapcar
	   (function (lambda (entry)
		       (if (stringp (car entry))
			   (progn
			     (setq header (mh-get-field (car entry)))
			     (if (zerop (length header))
				 ()
			       (mapcar
				(function (lambda (tuple)
					    (if (string-match (car tuple)
							      header)
						(throw 'match (cdr tuple)))))
				(cdr entry))))
			 (progn
			   (mapcar
			    ;; Here, we don't use "(let (result) ..." because
			    ;; we don't want the extra overhead.  Instead, we
			    ;; declare `result' in an outer `let' ....
			    (function (lambda (tuple)
					(if (setq result
						  (funcall (car tuple)))
					    (progn
					      (if (stringp result)
						  (throw 'match result)
						(throw 'match (cdr tuple)))))))
			    (cdr entry)))
			 )))
	   alist)
	  (if (eq operation (car mh-last-destination))
	      (symbol-name (cdr mh-last-destination))
	    ""))))))


(defun mh-get-default-refile-folder ()
  "Get a default name for refiling a message."
  (let ( (mh-current-folder mh-current-folder) )
    (if mh-auto-folder-alist
	(save-window-excursion
	  (save-excursion
	    (mh-show (mh-get-msg-num t))
	    (set-buffer mh-show-buffer)
	    (mh-auto-select-folder mh-auto-folder-alist 'refile)
	    ))
      (if (eq 'refile (car mh-last-destination))
	  (symbol-name (cdr mh-last-destination))
	"")
      )
    ))


(defun mh-get-the-from-entry ()
  "From mh-auto-folder-alist, get the list corresponding to the \"From:\" entry.
If there is none, return nil."
  (let (folders)
    (catch 'exit
      (setq folders mh-auto-folder-alist)
      (while folders
	(if (string= (car (car folders)) "From:")
	    (throw 'exit (cdr (car folders))))
	(setq folders (cdr folders))
	)
      (throw 'exit nil)
      )))


(defun mh-set-default-refile-folder (new-folder)
  "Set the default folder for the MH message under the cursor in the current
folder buffer."
  (interactive (list
		(mh-prompt-for-folder
		 "Set default refile"
		 (mh-get-default-refile-folder) t)))
  (let ( (from (mh-get-from-field))
	 folder-list folder previous new-folder-list new-from new-from-entry)
    (if (not from)
	(error "Unable to extract \"From:\" field"))
    (catch 'exit
      (progn
	(setq new-from (concat "^" (regexp-quote from) "$"))
	(if (setq folder-list (mh-get-the-from-entry))
	    (progn
	      ;; A "From:" entry exists
	      (while folder-list
		(if (string= (car (car folder-list)) new-from)
		    (progn
		      (setcdr (car folder-list) new-folder)
		      (throw 'exit nil)
		      ))
		(setq folder-list (cdr folder-list))
		)
	      ;; However, the "From:" entry does not already have a default
	      ;; entry for the new "from" address.  We must now add a new
	      ;; entry to be beginning of the list, and so we must rebuild
	      ;; the mh-auto-folder-alist variable.
	      (setq folder-list mh-auto-folder-alist)
	      (setq new-folder-list nil)
	      (while folder-list
		(if (string= (car (car folder-list)) "From:")
		    (progn
		      (setq new-from-entry
			    (list (cons "From:"
					(append (list (cons
						       new-from new-folder))
						(cdr (car folder-list))))))
		      (setq new-folder-list (append new-folder-list
						    new-from-entry))
		      )
		  (progn
		    (setq new-folder-list (append new-folder-list
						  (list (car folder-list))))
		    ))
		(setq folder-list (cdr folder-list))
		)
	      (setq mh-auto-folder-alist new-folder-list)
	      )
	  (progn
	    ;; A "From:" entry does not exist, so make one.
	    (setq mh-auto-folder-alist (cons (cons "From:"
						   (cons (cons new-from
							       new-folder)
							 nil))
					     mh-auto-folder-alist))
	    ))
	))
    (if (interactive-p)
	(message (format "Folder \"%s\" is now the default folder for \"%s\""
			 new-folder from)))
    ))


;; mh-e-vars-to-save is defined in mh-db.el, which is required by mh-e.el.
(if (not (memq 'mh-auto-folder-alist mh-e-vars-to-save))
    (setq mh-e-vars-to-save (cons 'mh-auto-folder-alist mh-e-vars-to-save)))
